#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  wb_jq.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
import pyjq
from jsonschema import validate

import os
import yaml
import gevent_inotifyx as inotify
from gevent.lock import Semaphore
from gevent import sleep
from glob import glob
from yaml.parser import ParserError


SCHEMA = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string"
        },
        "expression": {
            "type": "string"
        },
        "queue": {
            "type": "string"
        },
        "payload": {
            "type": "object",
            "patternProperties": {
                ".*": {
                    "type": [
                        "string",
                        "number"
                    ],
                }
            }
        },
    },
    "required": ["name", "expression", "queue"],
    "additionalProperties": False
}

SCHEMA_DISK = {
    "type": "object",
    "properties": {
        "expression": {
            "type": "string"
        },
        "queue": {
            "type": "string"
        },
        "payload": {
            "type": "object",
            "patternProperties": {
                ".*": {
                    "type": [
                        "string",
                        "number"
                    ],
                }
            }
        },
    },
    "required": ["expression", "queue"],
    "additionalProperties": False
}


class JQ(Actor):

    '''**Evaluates a data structure against jq expressions.**

    Evalutes (JSON) data structures against a set of jq expressions to decide
    which queue to forward the event to.

    JQ expressions
    --------------

    More information about jq expressions can be found here:

        - https://stedolan.github.io/jq/manual/


    Jq expressions need to return either **True** or **False**, otherwise this
    module will consider the result to be invalid and therefor skip the
    condition.

    Module level conditions
    -----------------------

    The module accepts the <conditions> parameter which is a list of
    conditions to evaluate against each data structure coming in.
    Each condition should have following format:

    JSON-schema::

        {
        "type": "object",
        "properties": {
            "name": {
                "type": "string"
            },
            "expression": {
                "type": "string"
            },
            "queue": {
                "type": "string"
            },
            "payload": {
                "type": "object",
                "patternProperties": {
                    ".*": {
                        "type": [
                            "string",
                            "number"
                        ],
                    }
                }
            },
        },
        "required": ["name", "expression", "queue"],
        "additionalProperties": False
        }


    Example::

        { "name": "test",
          "expression": ".greeting | test( "hello" )",
          "queue": "outbox",
          "payload": {
            "@tmp.some.key": 1,
          }
        }

    Disk level conditions
    ---------------------

    The directory <location> contains the conditions in YAML format. One
    condition is one file.  Files not having '.yaml' extension are ignored.

    This directory is monitored for changes and automatically reloaded
    whenever something changes.

    The rules should have following format:

    JSON-schema::

        {
        "type": "object",
        "properties": {
            "expression": {
                "type": "string"
            },
            "queue": {
                "type": "string"
            },
            "payload": {
                "type": "object",
                "patternProperties": {
                    ".*": {
                        "type": [
                            "string",
                            "number"
                        ],
                    }
                }
            },
        },
        "required": ["expression", "queue"],
        "additionalProperties": False
        }

    Example::

        queue: nagios
        expression: '.type | test( "nagios" )'

    payload
    -------

    The payload is a dictionary where keys are wishbone event references.


    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - conditions(dict)([])
           |  A dictionary consisting out of expression, queue, payload.

        - location(str)("")
           |  A directory containing rules.  This directory will be monitored
           |  for changes and automatically read for changes.
           |  An empty value disables this functionality.


    Queues:

        - inbox
           |  Incoming events.

        - no_match
           |  Events which did not match at least one rule.
    '''

    def __init__(self, actor_config, selection="@data", conditions=[], location=""):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("no_match")
        self.registerConsumer(self.consume, "inbox")
        self.disk_conditions = []
        self.condition_read_lock = Semaphore()

    def preHook(self):

        self.kwargs.conditions = self.validateConditions(self.kwargs.conditions)

        if self.kwargs.location == "":
            self.logging.info("No rules directory defined, not reading rules from disk.")
            self.kwargs.conditions = self.compileRules(self.kwargs.conditions)
        else:
            self.logging.info("Rules directoy '%s' defined." % (self.kwargs.location))
            self.monitor_location = ReadRulesDisk(self.logging, self.kwargs.location)
            self.disk_conditions = self.compileRules(self.monitor_location.readDirectory())
            self.logging.info("Read %s rules from disk and %s defined in config." % (len(self.disk_conditions), len(self.kwargs.conditions)))
            self.sendToBackground(self.monitorRuleDirectory)

    def compileRules(self, conditions):

        ok = []
        for condition in conditions:
            try:
                condition["compiled"] = pyjq.compile(condition["expression"])
            except Exception as err:
                message = err.message.replace("\n", " -> ")
                self.logging.error("Failed to compile jq rule '%s'. Skipped.  Reason: '%s'" % (condition["name"], message))
            else:
                ok.append(condition)
        return ok

    def monitorRuleDirectory(self):

        '''
        Loads new rules when changes happen.
        '''

        while self.loop():
            try:
                fresh_conditions = self.monitor_location.waitForChanges()
                with self.condition_read_lock:
                    if cmp(self.disk_conditions, fresh_conditions) != 0:
                        self.disk_conditions = self.compileRules(fresh_conditions)
                        self.logging.info("Read %s rules from disk and %s defined in config." % (len(self.disk_conditions), len(self.kwargs.conditions)))
            except Exception as err:
                self.logging.warning("Problem reading rules directory.  Reason: %s" % (err))
                sleep(1)

    def consume(self, event):

        matched = False
        data = event.get(self.kwargs.selection)
        with self.condition_read_lock:
            for condition in self.kwargs.conditions + self.disk_conditions:

                if self.pool.hasQueue(condition["queue"]):
                    try:
                        result = condition["compiled"].first(data)
                    except Exception as err:
                        self.logging.error("Skipped invalid jq expression '%s'. Reason: %s" % (condition["name"], err.message.replace("\n", " -> ")))
                        continue

                    if isinstance(result, bool) or result is None:
                        if result:
                            matched = True
                            e = event.clone()
                            e.set("@tmp.%s.rule" % (self.name), condition["name"])
                            if "payload" in condition:
                                for key, value in condition["payload"].iteritems():
                                    e.set(value, key)
                            self.submit(e, self.pool.getQueue(condition["queue"]))
                    else:
                        self.logging.error("Jq expression '%s' does not return a bool therefor it is skipped." % (condition["name"]))
                else:
                    self.logging.warning("Condition '%s' has queue '%s' defined but nothing connected." % (condition["name"], condition["queue"]))

        if not matched:
            self.submit(event, self.pool.queue.no_match)

    def validateConditions(self, conditions):
        '''
        Validates <conditions> whether each condition is valid.
        Returns a list of valid conditions.
        '''

        rules = []
        for condition in conditions:
            try:
                validate(condition, SCHEMA)
            except Exception as err:
                self.logging.error("Rule '%s' is invalid therefor skipped. Reason: %s" % (condition, err.message))
            finally:
                rules.append(condition)
        return rules


class ReadRulesDisk():

    '''
    Loads jq rules from a directory and monitors the directory for
    changes.

    Parameters:

        directory(string):   The directory to load rules from.
                            default: rules/

    '''

    def __init__(self, logger, directory="rules/"):
        self.logging = logger
        self.directory = directory

        if not os.access(self.directory, os.R_OK):
            raise Exception("Directory '%s' is not readable. Please verify." % (self.directory))

        self.fd = inotify.init()
        self.wb = inotify.add_watch(self.fd, self.directory, inotify.IN_CLOSE_WRITE + inotify.IN_CREATE + inotify.IN_DELETE + inotify.IN_MODIFY + inotify.IN_MOVE)
        self.logging.info("Monitoring directory '%s' for changes" % (directory))

    def createDir(self):

        if os.path.exists(self.kwargs.location):
            if not os.path.isdir(self.kwargs.location):
                raise Exception("%s exists but is not a directory" % (self.kwargs.location))
            else:
                self.logging.info("Directory %s exists so I'm using it." % (self.kwargs.location))
        else:
            self.logging.info("Directory %s does not exist so I'm creating it." % (self.kwargs.location))
            os.makedirs(self.kwargs.location)

    def readDirectory(self):

        return self.__readDirectory()

    def waitForChanges(self):

        inotify.get_events(self.fd)
        return self.readDirectory()

    def __readDirectory(self):
        '''Reads the content of the given directory and creates a dict
        containing the rules.'''

        rules = []
        for filename in glob("%s/*.yaml" % (self.directory)):
            try:
                with open(filename, 'r') as f:
                    rule_name = os.path.basename(filename).rstrip(".yaml")
                    rule = yaml.load("".join(f.readlines()))
                    try:
                        validate(rule, SCHEMA_DISK)
                    except Exception as err:
                        self.logging.warning("Rule %s not valid. Skipped. Reason: %s" % (filename, err.message))
                    else:
                        rule["name"] = rule_name
                        rules.append(rule)
            except ParserError as err:
                self.logging.warning("Failed to parse file %s.  Please validate the YAML syntax in a parser." % (filename))
            except IOError as err:
                self.logging.warning("Failed to read %s.  Reason: %s" % (filename, err))
            except Exception as err:
                self.logging.warning("Unknown error parsing file %s.  Skipped.  Reason: %s." % (filename, err))

        return rules




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  setup.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys
import os

# Required to make pycrypto compile with PyPy
# https://github.com/dlitz/pycrypto/pull/59
if "PyPy" in sys.version:
    print("We are running PyPY, disable gmp to prevent pycrypto build error.")
    os.environ["with_gmp"] = "no"

PROJECT = 'wishbone'
VERSION = '2.1.1'

install_requires = ['gevent==1.1rc4',
                    'greenlet==0.4.9',
                    'argparse==1.3.0',
                    'prettytable==0.7.2',
                    'python-daemon==1.6',
                    'pyyaml==3.11',
                    'msgpack-python==0.4.7',
                    'pyzmq==14.7.0',
                    'amqp==1.4.9',
                    'jinja2==2.8',
                    'jsonschema==2.5.1',
                    'gearman==2.0.2',
                    'pycrypto==2.6.1',
                    'flask==0.10.1',
                    'gevent_inotifyx==0.1.1',
                    'requests==2.7.0',
                    'colorama==0.3.3',
                    'arrow==0.6.0',
                    'elasticsearch==1.6.0',
                    'importlib==1.0.3',
                    'uplook==0.4.1',
                    'pyjq==1.1']

# Dirty hack to make readthedocs build the docs
# For some reason mocking out jq as documented is not working

if os.environ.get("READTHEDOCS", False):
    dependency_links = []
    install_requires.remove('pyjq==1.1')

else:
    dependency_links = [
        'https://github.com/smetj/pyjq/tarball/master#egg=pyjq-1.1'
    ]

# Deps pulled in by other modules
# 'lockfile==0.10.2'
# 'MarkupSafe==0.23'
# 'repoze.lru==0.6',
# 'werkzeug==0.10.4',
# 'itsdangerous==0.24',
# 'inotifyx==0.2.2',
# 'six==1.9.0',
# 'python-dateutil==2.4.2',
# 'urllib3==1.12',

try:
    with open('README.rst', 'rt') as f:
        long_description = f.read()
except IOError:
    long_description = ''


class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)

setup(
    name=PROJECT,
    version=VERSION,

    description='Build composable event pipeline servers with minimal effort.',
    long_description=long_description,

    author='Jelle Smet',
    author_email='development@smetj.net',

    url='https://github.com/smetj/wishbone',
    download_url='https://github.com/smetj/wishbone/tarball/master',
    dependency_links=dependency_links,
    classifiers=['Development Status :: 5 - Production/Stable',
                 'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: Implementation :: PyPy',
                 'Intended Audience :: Developers',
                 'Intended Audience :: System Administrators',
                 ],
    extras_require={
        'testing': ['pytest'],
    },
    platforms=['Linux'],
    test_suite='tests.test_wishbone',
    cmdclass={'test': PyTest},
    scripts=[],
    provides=[],
    install_requires=install_requires,
    namespace_packages=[],
    packages=find_packages(),
    package_data={'': ['data/wordlist.txt', 'data/LICENCE', 'data/sse.html', 'data/banner.tmpl']},
    zip_safe=False,
    entry_points={
        'console_scripts': ['wishbone = wishbone.bootstrap:main'],
        'wishbone.flow': [
            'fanout = wishbone.module.fanout:Fanout',
            'funnel = wishbone.module.funnel:Funnel',
            'match = wishbone.module.match:Match',
            'roundrobin = wishbone.module.roundrobin:RoundRobin',
            'ttl = wishbone.module.ttl:TTL',
            'jsonvalidate = wishbone.module.jsonvalidate:JSONValidate',
            'jq = wishbone.module.wb_jq:JQ'
        ],
        'wishbone.encode': [
            'graphite = wishbone.module.graphite:Graphite',
            'humanlogformat = wishbone.module.humanlogformat:HumanLogFormat',
            'influxdb = wishbone.module.influxdb:InfluxDB',
            'msgpack = wishbone.module.msgpackencode:MSGPackEncode',
            'json = wishbone.module.jsonencode:JSONEncode'
        ],
        'wishbone.decode': [
            'msgpack = wishbone.module.msgpackdecode:MSGPackDecode',
            'json = wishbone.module.jsondecode:JSONDecode'
        ],
        'wishbone.function': [
            'deserialize = wishbone.module.deserialize:Deserialize',
            'modify = wishbone.module.modify:Modify',
            'loglevelfilter = wishbone.module.loglevelfilter:LogLevelFilter',
            'template = wishbone.module.template:Template'
        ],
        'wishbone.input': [
            'amqp = wishbone.module.amqpin:AMQPIn',
            'dictgenerator = wishbone.module.dictgenerator:DictGenerator',
            'disk = wishbone.module.diskin:DiskIn',
            'fresh = wishbone.module.fresh:Fresh',
            'httpclient = wishbone.module.httpinclient:HTTPInClient',
            'httpserver = wishbone.module.httpinserver:HTTPInServer',
            'namedpipe = wishbone.module.namedpipein:NamedPipeIn',
            'tcp = wishbone.module.tcpin:TCPIn',
            'testevent = wishbone.module.testevent:TestEvent',
            'udp = wishbone.module.udpin:UDPIn',
            'gearman = wishbone.module.gearmanin:GearmanIn',
            'zeromq_topic = wishbone.module.zmqtopicin:ZMQTopicIn',
            'zeromq_pull = wishbone.module.zmqpullin:ZMQPullIn'
        ],
        'wishbone.output': [
            'amqp = wishbone.module.amqpout:AMQPOut',
            'disk = wishbone.module.diskout:DiskOut',
            'elasticsearch = wishbone.module.elasticsearchout:ElasticSearchOut',
            'email = wishbone.module.emailout:EmailOut',
            'file = wishbone.module.fileout:FileOut',
            'http = wishbone.module.httpoutclient:HTTPOutClient',
            'null = wishbone.module.null:Null',
            'stdout = wishbone.module.stdout:STDOUT',
            'syslog = wishbone.module.wbsyslog:Syslog',
            'tcp = wishbone.module.tcpout:TCPOut',
            'udp = wishbone.module.udpout:UDPOut',
            'uds = wishbone.module.udsout:UDSOut',
            'sse = wishbone.module.sse:ServerSentEvents',
            'zeromq_topic = wishbone.module.zmqtopicout:ZMQTopicOut',
            'zeromq_push = wishbone.module.zmqpushout:ZMQPushOut'
        ],
        'wishbone.contrib.flow': [
        ],
        'wishbone.contrib.encode': [
        ],
        'wishbone.contrib.decode': [
        ],
        'wishbone.contrib.function': [
        ],
        'wishbone.contrib.input': [
        ],
        'wishbone.contrib.output': [
        ]
    }
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  tcpin.py
#
#  Copyright 2016 Jelle Smet development@smetj.net
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Event
from gevent.server import StreamServer
from gevent.pool import Pool
from gevent import socket, sleep


class TCPIn(Actor):

    '''**A TCP server.**

    Creates a TCP socket to which data can be submitted.

    Parameters:

        - address(str)("0.0.0.0")
           |  The address to bind to.

        - port(int)(19283)
           |  The port to bind to.

        - delimiter(str)("\\n")
           |  The delimiter which separates multiple
           |  messages in a stream of data.

        - max_connections(int)(0)
           |  The maximum number of simultaneous
           |  connections.  0 means "unlimited".

        - reuse_port(bool)(False)
           |  Whether or not to set the SO_REUSEPORT
           |  socket option.  Interesting when starting
           |  multiple instances and allow them to bind
           |  to the same port.
           |  Requires Linux kernel >= 3.9

    Queues:

        - outbox
           |  Data coming from the outside world.


    **delimiter**

    When no delimiter is defined, all incoming data between connect and
    disconnect is considered to be 1 event. When a delimiter is defined,
    multiple events are extracted out of a data stream using the defined
    delimiter.  The module will check each line of data whether it ends with
    the delimiter.  If not the line will be added to an internal buffer.  If
    so, the delimiter will be stripped of and when there is data left, it will
    be added to the buffer after which the buffer will be flushed as one
    Wishbone message/event.  The advantage is that a client can stay connected
    and stream data.

    Choosing "\\n" as a delimiter, which is the default, each new line is a new event.

    '''

    def __init__(self, actor_config, port=19283, address='0.0.0.0', delimiter="\n", max_connections=0, reuse_port=False):
        Actor.__init__(self, actor_config)

        if self.kwargs.delimiter is None:
            self.handle = self.__handleNoDelimiter
        elif self.kwargs.delimiter == "\n":
            self.handle = self.__handleNextLine
        else:
            self.handle = self.__handleDelimiter

        self.pool.createQueue("outbox")

    def preHook(self):
        self.sock = self.__setupSocket(self.kwargs.address, self.kwargs.port)
        self.logging.info("TCP server initialized on address %s and port %s." % (self.kwargs.address, self.kwargs.port))
        self.sendToBackground(self.serve)

    def __setupSocket(self, address, port):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if self.kwargs.reuse_port:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.setsockopt(socket.SOL_SOCKET, 15, 1)
        else:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        sock.setblocking(0)
        sock.bind((address, port))
        sock.listen(1)
        return sock

    def __handleNoDelimiter(self, sock, address):
        sfile = sock.makefile()
        chunk = sfile.readlines()
        event = self.createEvent()
        event.data = ''.join(chunk)
        self.submit(event, self.pool.queue.outbox)
        sfile.close()
        sock.close()

    def __handleNextLine(self, sock, address):
        self.logging.debug("Connection from %s." % (str(address[0])))
        sfile = sock.makefile()

        while self.loop():
            try:
                chunk = sfile.readline()
            except:
                self.logging.debug("Client %s disconnected." % (str(address[0])))
                break

            if not chunk:
                self.logging.debug("Client %s disconnected." % (str(address[0])))
                break
            else:
                event = Event(chunk.rstrip('\r\n'))
                self.submit(event, self.pool.queue.outbox)

    def __handleDelimiter(self, sock, address):
        self.logging.debug("Connection from %s." % (str(address[0])))
        sfile = sock.makefile()
        data = []

        while self.loop():
            chunk = sfile.readline()
            if not chunk:
                if len(data) > 0:
                    event = Event(''.join(data))
                    self.submit(event, self.pool.queue.outbox)
                self.logging.debug("Client %s disconnected." % (str(address[0])))
                break

            elif chunk.endswith(self.kwargs.delimiter):
                chunk = chunk.rstrip(self.kwargs.delimiter)
                if chunk != '':
                    data.append(chunk)
                    event = Event(''.join(data))
                    self.submit(event, self.pool.queue.outbox)
                    data = []
            else:
                data.append(chunk)

        sfile.close()
        sock.close()

    def serve(self):
        if self.kwargs.max_connections > 0:
            pool = Pool(self.kwargs.max_connections)
            self.logging.debug("Setting up a connection pool of %s connections." % (self.kwargs.max_connections))
            self.stream_server = StreamServer(self.sock, self.handle, spawn=pool)
            self.stream_server.start()
        else:
            self.stream_server = StreamServer(self.sock, self.handle)
            self.stream_server.start()
        while self.loop():
            sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  readrulesdisk.py
#
#  Copyright 2014 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from gevent import spawn
import gevent_inotifyx as inotify
from gevent import event
from glob import glob
import os
import yaml
from yaml.parser import ParserError

class ReadRulesDisk():

    '''
    Loads PySeps rules from a directory and monitors the directory for
    changes.

    Parameters:

        directory(string):   The directory to load rules from.
                            default: rules/

    '''

    def __init__(self, logger, directory="rules/"):
        self.logger = logger
        self.directory = directory

        if not os.access(self.directory, os.R_OK):
            raise Exception("Directory '%s' is not readable. Please verify." % (self.directory))

        self.fd = inotify.init()
        self.wb = inotify.add_watch(self.fd, self.directory, inotify.IN_CLOSE_WRITE + inotify.IN_CREATE + inotify.IN_DELETE + inotify.IN_MODIFY + inotify.IN_MOVE)

    def readDirectory(self):

        return self.__readDirectory()

    def waitForChanges(self, grace_period=5):

        events = inotify.get_events(self.fd)
        return self.readDirectory()

    def __readDirectory(self):
        '''Reads the content of the given directory and creates a dict
        containing the rules.'''

        rules = {}
        for filename in glob("%s/*.yaml" % (self.directory)):
            try:
                with open(filename, 'r') as f:
                    key_name = os.path.basename(filename).rstrip(".yaml")
                    rule = yaml.load("".join(f.readlines()))
                    try:
                        self.ruleCompliant(rule)
                    except Exception as err:
                        self.logger.warning("Rule %s not valid. Skipped. Reason: %s" % (filename, err))
                    else:
                        rules[key_name] = rule
            except ParserError as err:
                self.logger.warning("Failed to parse file %s.  Please validate the YAML syntax in a parser." % (filename))
            except IOError as err:
                self.logger.warning("Failed to read %s.  Reason: %s" % (filename, err))
            except Exception as err:
                self.logger.warning("Unknown error parsing file %s.  Skipped.  Reason: %s." % (filename, err))

        return rules

    def ruleCompliant(self, rule):

        '''Does basic rule validation'''

        assert isinstance(rule["condition"], list), "Condition needs to be of type list."
        for c in rule["condition"]:
            assert isinstance(c, dict), "An individual condition needs to be of type dict."
        assert isinstance(rule["queue"], list), "Queue needs to be of type list."


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  match.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone import Actor
from gevent import sleep
from .matchrules import MatchRules
from .readrules import ReadRulesDisk
import os
from copy import deepcopy
from gevent.lock import Semaphore


class Match(Actor):

    '''**Pattern matching on a key/value document stream.**

    This module routes messages to a queue associated to the matching rule
    set.  The '@data' payload has to be of <type 'dict'>.  Typically,
    the source data is JSON converted to a Python dictionary.

    The match rules can be either stored on disk or directly defined into the
    bootstrap file.

    A match rule is written in YAML syntax and consists out of 2 parts:

    - condition:

        A list of dictionaries holding with the individual conditions which
        ALL have to match for the complete rule to match.

    ::

        re:     Regex matching
        !re:    Negative regex matching
        >:      Bigger than
        >=:     Bigger or equal than
        <:      Smaller than
        <=:     Smaller or equal than
        =:      Equal than
        in:     Evaluate list membership
        !in:    Evaluate negative list membership


    - queue:

        The queue section contains a list of dictionaries/maps each containing
        1 key with another dictionary/map as a value.  These key/value pairs
        are added to the *header section* of the event and stored under the
        queue name key.
        If you are not interested in adding any information to the header you
        can leave the dictionary empty.  So this would be valid:

    All rules will be evaluated sequentially in no particular order.  When a
    rule matches, evaluation the other rules will continue untill all rules
    are processed.

    *Examples*

    This example would route the events - with field "greeting" containing
    the value "hello" - to the outbox queue without adding any information
    to the header of the event itself.

    ::

        condition:
            - greeting: re:^hello$

        queue:
            - outbox:



    This example combines multiple conditions and stores 4 variables under
    @tmp.<self.name> while submitting the event to the modules'
    **email** queue.

    ::

        condition:
            - check_command: re:check:host.alive
            - hostproblemid: re:\d*
            - hostgroupnames: in:tag:development

        queue:
            - email:
                from: monitoring@yourdomain.com
                to:
                    - oncall@yourdomain.com
                subject: UMI - Host  {{ hostname }} is  {{ hoststate }}.
                template: host_email_alert



    Parameters:

        - name(str)
           |  The name of the module.

        - size(int)
           |  The default max length of each queue.

        - frequency(int)
           |  The frequency in seconds to generate metrics.

        - location(str)("")
           |  The directory containing rules.
           |  If empty, no rules are read from disk.

        - rules(dict)({})
           |  A dict of rules in the above described format.
           |  For example:
           |  {"omg": {"condition": [{"greeting": "re:^hello$"}], "queue": [{"outbox": {"one": 1}}]}}

        - ignore_missing_fields(bool)(False)
           |  When a doc is missing a field which is evaluated in the
           |  condition this will simply be ignored and therefor can still yield a match.
           |  When set to False(default) a missing field will automatically result in a non-match.


    Queues:

        - inbox
           |  Incoming events

        - <queue_name>
           |  The queue which matches a rule

        - nomatch
           |  The queue receiving event without matches

    '''

    def __init__(self, actor_config, location="", rules={}, ignore_missing_fields=False):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("nomatch")
        self.registerConsumer(self.consume, "inbox")

        self.__active_rules = {}
        self.match = MatchRules()
        self.rule_lock = Semaphore()

    def preHook(self):

        if self.kwargs.location != "":
            self.createDir()
            self.logging.info("Rules directoy '%s' defined." % (self.kwargs.location))
            self.sendToBackground(self.monitorRuleDirectory)
        else:
            self.__active_rules.update(self.uplook.dump()["rules"])
            self.logging.info("No rules directory defined, not reading rules from disk.")

    def createDir(self):

        if os.path.exists(self.kwargs.location):
            if not os.path.isdir(self.kwargs.location):
                raise Exception("%s exists but is not a directory" % (self.kwargs.location))
            else:
                self.logging.info("Directory %s exists so I'm using it." % (self.kwargs.location))
        else:
            self.logging.info("Directory %s does not exist so I'm creating it." % (self.kwargs.location))
            os.makedirs(self.kwargs.location)

    def monitorRuleDirectory(self):

        '''
        Loads new rules when changes happen.
        '''

        self.logging.info("Monitoring directory '%s' for changes" % (self.kwargs.location))
        self.readRules = ReadRulesDisk(self.logging, self.kwargs.location)

        new_rules = self.readRules.readDirectory()
        new_rules.update(self.kwargs.rules)
        self.__active_rules = new_rules
        self.logging.info("Read %s rules from disk and %s defined in config." % (len(new_rules), len(self.kwargs.rules)))
        while self.loop():
            try:
                new_rules = self.readRules.waitForChanges()
                new_rules.update(self.kwargs.rules)
                if cmp(self.__active_rules, new_rules) != 0:
                    self.rule_lock.acquire()
                    self.__active_rules = new_rules
                    self.rule_lock.release()
                    self.logging.info("Read %s rules from disk and %s defined in config." % (len(new_rules), len(self.kwargs.rules)))
            except Exception as err:
                self.logging.warning("Problem reading rules directory.  Reason: %s" % (err))
                sleep(1)

    def consume(self, event):
        '''Submits matching documents to the defined queue along with
        the defined header.'''

        if isinstance(event.get(), dict):
            self.rule_lock.acquire()
            for rule in self.__active_rules:
                e = deepcopy(event)
                if self.evaluateCondition(self.__active_rules[rule]["condition"], e.get()):
                    e.set(rule, '@tmp.%s.rule' % (self.name))
                    for queue in self.__active_rules[rule]["queue"]:
                        event_copy = deepcopy(e)
                        for name in queue:
                            if queue[name] is not None:
                                for key, value in queue[name].iteritems():
                                    event_copy.set(value, '@tmp.%s.%s' % (self.name, key))
                            self.submit(event_copy, self.pool.getQueue(name))
                else:
                    e.set(rule, "@tmp.%s.rule" % (self.name))
                    self.submit(e, self.pool.queue.nomatch)
                    # self.logging.debug("No match for rule %s." % (rule))
            self.rule_lock.release()
        else:
            raise Exception("Incoming data is not of type dict, dropped.")

    def evaluateCondition(self, conditions, fields):
        for condition in conditions:
            for field in condition:
                if field in fields:
                    try:
                        match_result = self.match.do(condition[field], fields[field])
                    except Exception as err:
                        self.logging.error("Invalid condition '%s'. Skipped.  Reason: %s" % (condition[field], err))
                        return False
                    finally:
                        if not match_result:
                            # self.logging.debug("field %s with condition %s DOES NOT MATCH value %s" % (field, condition[field], fields[field]))
                            return False
                else:
                    if not self.kwargs.ignore_missing_fields:
                        return False
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  zmqpullin.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone import Actor
from wishbone.event import Event
import zmq.green as zmq


class ZMQPullIn(Actor):

    '''**Pulls events from one or more ZeroMQ push modules.**

    Expects to connect with one or more wishbone.input.push modules.  This
    module can be started in client or server mode.  In server mode, it waits
    for incoming connections.  In client mode it connects to the defined
    servers.  Events are spread in a round robin pattern over all connections.

    Parameters:

        - mode(str)("server")
           |  The mode to run in.  Possible options are:
           |  - server: Binds to a port and listens.
           |  - client: Connects to a port.

        - interface(string)("0.0.0.0")
           |  The interface to bind to in server <mode>.

        - port(int)(19283)
           |  The port to bind to in server <mode>.

        - servers(list)([])
           |  A list of hostname:port entries to connect to.
           |  Only valid when running in "client" <mode>.


    Queues:

        - outbox
           |  Events arriving from the outside.

    '''

    def __init__(self, actor_config, mode="server", interface="0.0.0.0", port=19283, servers=[]):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("outbox")

    def preHook(self):

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PULL)

        if self.kwargs.mode == "server":
            self.socket.bind("tcp://*:%s" % self.kwargs.port)
            self.logging.info("Listening on port %s" % (self.kwargs.port))
        else:
            self.socket.connect("tcp://%s" % self.kwargs.servers[0])

        self.sendToBackground(self.drain)

    def drain(self):

        while self.loop():
            data = self.socket.recv()
            event = Event(data)
            self.submit(event, self.pool.queue.outbox)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  httpoutclient.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from gevent import monkey; monkey.patch_socket()
from wishbone import Actor
import requests


class HTTPOutClient(Actor):

    '''**Posts data to the requested URL**

    Posts data to a remote URL


    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - method(str)("PUT")
           |  The http method to use. PUT/POST

        - content_type(str)("application/json")
           |  The content type to use.

        - accept(str)("text/plain")
           |  The accept value to use.

        - url(str)("http://localhost")
           |  The url to submit the data to

        - username(str)
           |  The username to authenticate

        - password(str)
           |  The password to authenticate


    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, config, selection="@data", method="PUT", content_type="application/json", accept="text/plain", url="https://localhost", username=None, password=None):

        Actor.__init__(self, config)
        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        if self.kwargs.method == "PUT":
            self.submitToResource = self.__put
        elif self.kwargs.method == "POST":
            self.submitToResource = self.__post
        else:
            raise Exception("Invalid http method defined: '%s'." % self.kwargs.method)

        if self.kwargs.url.startswith('https'):
            monkey.patch_ssl()

    def consume(self, event):

        try:
            response = self.submitToResource(event.get(self.kwargs.selection))
            response.raise_for_status()
        except Exception as err:
            self.logging.error("Failed to submit data.  Reason: %s" % (err))
            event.set(str(response.text), "@tmp.%s.server_response" % (self.name))
            raise

    def __put(self, data):

        return requests.put(self.kwargs.url, data=data, auth=(self.kwargs.username, self.kwargs.password), headers={'Content-type': self.kwargs.content_type, 'Accept': self.kwargs.accept})

    def __post(self, data):

        return requests.post(self.kwargs.url, data=data, auth=(self.kwargs.username, self.kwargs.password), headers={'Content-type': self.kwargs.content_type, 'Accept': self.kwargs.accept})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  amqpout.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from gevent import monkey
monkey.patch_socket()
from wishbone import Actor
from amqp.connection import Connection as amqp_connection
from amqp import basic_message
from gevent import sleep


class AMQPOut(Actor):

    '''**Produces messages to AMQP.**

    Submits messages to an AMQP message broker.

    If <exchange> is not provided, no exchange will be created during initialisation.
    If <queue> is not provided, queue will be created during initialisation

    If <exchange> and <queue> are provided, they will both be created and
    bound during initialisation.

    <exchange> and <queue> can be event lookup values.

    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - host(str)("localhost")
           |  The host broker to connect to.

        - port(int)(5672)
           |  The port to connect to.

        - vhost(str)("/")
           |  The virtual host to connect to.

        - user(str)("guest")
           |  The username to authenticate.

        - password(str)("guest")
           |  The password to authenticate.

        - exchange(str)("")
           |  The exchange to declare.

        - exchange_type(str)("direct")
           |  The exchange type to create. (direct, topic, fanout)

        - exchange_durable(bool)(false)
           |  Declare a durable exchange.

        - queue(str)("")
           |  The queue to declare and bind to <exchange>. This will also the
           |  the destination queue of the submitted messages unless
           |  <routing_key> is set to another value and <exchange_type> is
           |  "topic".

        - queue_durable(bool)(false)
           |  Declare a durable queue.

        - queue_exclusive(bool)(false)
           |  Declare an exclusive queue.

        - queue_auto_delete(bool)(true)
           |  Whether to autodelete the queue.

        - queue_lazy(bool)(false)
            |  When true <queue> is a lazy queue.

        - routing_key(str)("")
           |  The routing key to use when submitting messages.

        - delivery_mode(int)(1)
           |  Sets the delivery mode of the messages.


    Queues:

        - inbox
           | Messages going to the defined broker.
    '''

    def __init__(self, actor_config, selection="@data",
                 host="localhost", port=5672, vhost="/", user="guest", password="guest",
                 exchange="", exchange_type="direct", exchange_durable=False,
                 queue="", queue_durable=False, queue_exclusive=False, queue_auto_delete=True,
                 routing_key="", delivery_mode=1, queue_lazy=False):

        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):
        self.__arguments = {}
        if self.kwargs.queue_lazy:
            self.__arguments["x-queue-mode"] = "lazy"
        self.sendToBackground(self.setupConnectivity)

    def consume(self, event):

        data = str(event.get(self.kwargs.selection))

        message = basic_message.Message(
                    body=data,
                    delivery_mode=self.kwargs.delivery_mode
                    )

        self.channel.basic_publish(message,
                                   exchange=self.kwargs.exchange,
                                   routing_key=self.kwargs.routing_key)
        sleep(0)

    def setupConnectivity(self):

        while self.loop():
            try:
                self.connection = amqp_connection(host=self.kwargs.host,
                                                  port=self.kwargs.port,
                                                  virtual_host=self.kwargs.vhost,
                                                  userid=self.kwargs.user,
                                                  password=self.kwargs.password)

                self.channel = self.connection.channel()

                if self.kwargs.exchange != "":
                    self.channel.exchange_declare(
                        self.kwargs.exchange,
                        self.kwargs.exchange_type,
                        durable=self.kwargs.exchange_durable
                    )
                    self.logging.debug("Declared exchange %s." % (self.kwargs.exchange))

                if self.kwargs.queue != "":
                    self.channel.queue_declare(
                        self.kwargs.queue,
                        durable=self.kwargs.queue_durable,
                        exclusive=self.kwargs.queue_exclusive,
                        auto_delete=self.kwargs.queue_auto_delete,
                        arguments=self.__arguments
                    )
                    self.logging.debug("Declared queue %s." % (self.kwargs.queue))

                if self.kwargs.exchange != "" and self.kwargs.queue != "":
                    self.channel.queue_bind(
                        self.kwargs.queue,
                        self.kwargs.exchange,
                        routing_key=self.kwargs.routing_key
                    )
                    self.logging.debug("Bound queue %s to exchange %s." % (self.kwargs.queue, self.kwargs.exchange))

                self.logging.info("Connected to broker.")
                break
            except Exception as err:
                self.logging.error("Failed to connect to broker.  Reason %s " % (err))
                sleep(1)

    def postHook(self):
        try:
            self.connection.close()
        except:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  amqpin.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from gevent import monkey;monkey.patch_all()
from wishbone import Actor
from amqp.connection import Connection as amqp_connection
from gevent import sleep
from wishbone.event import Event


class AMQPIn(Actor):

    '''**Consumes messages from AMQP.**

    Consumes messages from an AMQP message broker.
    The declared <exchange> and <queue> will be bound to each other.

    Parameters:

        - host(str)("localhost")
           | The host to connect to.

        - port(int)(5672)
           | The port to connect to.

        - vhost(str)("/")
           |  The virtual host to connect to.

        - user(str)("guest")
           |  The username to authenticate.

        - password(str)("guest")
           |  The password to authenticate.

        - exchange(str)("")
           |  The exchange to declare.

        - exchange_type(str)("direct")
           |  The exchange type to create. (direct, topic, fanout)

        - exchange_durable(bool)(false)
           |  Declare a durable exchange.

        - exchange_auto_delete(bool)(true)
           |  If set, the exchange is deleted when all queues have finished using it.

        - exchange_passive(bool)(false)
           |  If set, the server will not create the exchange. The client can use
           |  this to check whether an exchange exists without modifying the server state.

        - queue(str)("wishbone")
           |  The queue to declare and ultimately consume.

        - queue_durable(bool)(false)
           |  Declare a durable queue.

        - queue_exclusive(bool)(false)
           |  Declare an exclusive queue.

        - queue_auto_delete(bool)(true)
           |  Whether to autodelete the queue.

        - queue_declare(bool)(true)
           |  Whether to actually declare the queue.

        - queue_lazy(bool)(false)
            |  When true <queue> is a lazy queue.

        - routing_key(str)("")
           |  The routing key to use in case of a "topic" exchange.
           | When the exchange is type "direct" the routing key is always equal
           | to the <queue> value.

        - prefetch_count(int)(1)
           |  Prefetch count value to consume messages from queue.

        - no_ack(bool)(false)
           |  Override acknowledgement requirement.


    Queues:

        - outbox
           |  Messages coming from the defined broker.

        - ack
           |  Messages to acknowledge (requires the delivery_tag)

        - cancel
           |  Cancels a message acknowledgement (requires the delivery_tag)
    '''

    def __init__(self, actor_config, host="localhost", port=5672, vhost="/", user="guest", password="guest",
                 exchange="", exchange_type="direct", exchange_durable=False, exchange_auto_delete=True, exchange_passive=False,
                 queue="wishbone", queue_durable=False, queue_exclusive=False, queue_auto_delete=True, queue_declare=True,
                 routing_key="", prefetch_count=1, no_ack=False, queue_lazy=False):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("outbox")
        self.pool.createQueue("ack")
        self.pool.createQueue("cancel")
        self.pool.queue.ack.disableFallThrough()
        self.connection = None

    def preHook(self):
        self.__arguments = {}
        if self.kwargs.queue_lazy:
            self.__arguments["x-queue-mode"] = "lazy"
        self.sendToBackground(self.drain)
        self.sendToBackground(self.handleAcknowledgements)
        self.sendToBackground(self.handleAcknowledgementsCancel)

    def consume(self, message):
        event = Event(str(message.body))
        event.set(message.delivery_info["delivery_tag"], "@tmp.%s.delivery_tag" % (self.name))
        self.submit(event, self.pool.queue.outbox)

    def setupConnectivity(self):

        while self.loop():
            try:
                self.connection = amqp_connection(
                                    host=self.kwargs.host,
                                    port=self.kwargs.port,
                                    virtual_host=self.kwargs.vhost,
                                    userid=self.kwargs.user,
                                    password=self.kwargs.password
                                    )
                self.channel = self.connection.channel()

                if self.kwargs.exchange != "":
                    self.channel.exchange_declare(
                        self.kwargs.exchange,
                        self.kwargs.exchange_type,
                        durable=self.kwargs.exchange_durable,
                        auto_delete=self.kwargs.exchange_auto_delete,
                        passive=self.kwargs.exchange_passive
                    )
                    self.logging.debug("Declared exchange %s." % (self.kwargs.exchange))

                if self.kwargs.queue_declare:
                    self.channel.queue_declare(
                        self.kwargs.queue,
                        durable=self.kwargs.queue_durable,
                        exclusive=self.kwargs.queue_exclusive,
                        auto_delete=self.kwargs.queue_auto_delete,
                        arguments=self.__arguments
                    )
                    self.logging.debug("Declared queue %s." % (self.kwargs.queue))

                if self.kwargs.exchange != "":
                    self.channel.queue_bind(
                        self.kwargs.queue,
                        self.kwargs.exchange,
                        routing_key=self.kwargs.routing_key
                    )
                    self.logging.debug("Bound queue %s to exchange %s." % (self.kwargs.queue, self.kwargs.exchange))

                self.channel.basic_qos(prefetch_size=0, prefetch_count=self.kwargs.prefetch_count, a_global=False)
                self.channel.basic_consume(self.kwargs.queue, callback=self.consume, no_ack=self.kwargs.no_ack)
                self.logging.info("Connected to broker.")
            except Exception as err:
                self.logging.error("Failed to connect to broker.  Reason %s " % (err))
                sleep(1)
            else:
                break

    def drain(self):

        self.setupConnectivity()
        while self.loop():
            try:
                self.connection.drain_events()
                sleep(0)
            except Exception as err:
                self.logging.error("Problem connecting to broker.  Reason: %s" % (err))
                self.setupConnectivity()
                sleep(1)

    def handleAcknowledgements(self):
        while self.loop():
            event = self.pool.queue.ack.get()
            try:
                self.channel.basic_ack(event.get("@tmp.%s.delivery_tag" % (self.name)))
            except Exception as err:
                self.pool.queue.ack.rescue(event)
                self.logging.error("Failed to acknowledge message.  Reason: %s." % (err))
                sleep(0.5)

    def handleAcknowledgementsCancel(self):
        while self.loop():
            event = self.pool.queue.cancel.get()
            try:
                self.channel.basic_reject(event.get("@tmp.%s.delivery_tag" % (self.name)), True)
            except Exception as err:
                self.pool.queue.ack.rescue(event)
                self.logging.error("Failed to cancel acknowledge message.  Reason: %s." % (err))
                sleep(0.5)

    def postHook(self):
        try:
            self.connection.close()
        except:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  msgpackencode.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
import msgpack


class MSGPackEncode(Actor):
    '''
    **Encodes Python objects to MSGPack format.**

    Encodes the payload or complete events to MSGPack format.

    Parameters:

        - complete(bool)(False)
           |  When True encodes the complete event.  If False only
           |  encodes the data part.

    Queues:

        - inbox
           |  Incoming messages

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, complete=False):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")

        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        if self.kwargs.complete:
            self.encode = self.__encodeComplete
        else:
            self.encode = self.__encodeData

    def consume(self, event):
        event.set(self.encode(event))
        self.submit(event, self.pool.queue.outbox)

    def __encodeComplete(self, event):
        return msgpack.packb(event.dump(convert_timestamp=False))

    def __encodeData(self, event):
        return msgpack.packb(event.get())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  wbsyslog.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Log
import syslog
import sys
import os


class Syslog(Actor):

    '''**Writes log events to syslog.**

    Logevents have following format:

    (6, 1367682301.430527, 'Router', 'Received SIGINT. Shutting down.')

    The first value corresponds to the syslog severity level.

    Parameters:

        - selection(str)("@data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - level(int)(5)
           |  The loglevel.

        - ident(str)(<script_name>)
           |  The syslog id string.
           |  If not provided the script name is used.

    Queues:

        - inbox
           |  incoming events
    '''

    def __init__(self, actor_config, selection="@data", level=5, ident=os.path.basename(sys.argv[0])):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        syslog.openlog("%s[%s]" % (self.kwargs.ident, os.getpid()))

    def consume(self, event):

        d = event.get(self.kwargs.selection)
        if isinstance(d, Log):
            syslog.syslog(d.level, "%s: %s" % (d.module, d.message))
        else:
            syslog.syslog(self.kwargs.level, "%s: %s" % (self.kwargs.ident, str(d)))

    def postHook(self):
        syslog.closelog()
